"""
Program  : SDMinP
Version  : 1.0
Date     : 16.03.2005
Purpose  : Adjustment of the multiple type I error (FWER)
Programer: Markus Obreiter
Contact  : m.obreiter@dkfz-heidelberg.de
Copyright (C) 2005 Deutsches Krebsforschungszentrum, DKFZ (German Cancer
Research Center), Division of Clinical
Epidemiology

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your option)
any later version. 
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

A copy of the GNU Public License is attached to the program documentation.
"""

import os, sys, time, string
import inputDataFile, markerListHandler, formatHelper, logfile
import calculationRawP, calculationStepDownMinP, calculationGlobalPvalue
import resultfile

class Control:
    """this class controls the calculations"""

    def __init__(self, appCfgObject):

        self.appConfig    = appCfgObject
        self.DEBUG        = 0
        self.LOGTIME      = 0
        self.PROMPTRESULT = 0
        self.DISPLAY      = 1
        self.DISPLAYLEVEL = 0
        self.logtimePath = '../result/logtime.txt'        
        self.inputDataFile = inputDataFile.InputDataFile(self.appConfig)
        self.pValRawOrgIncluded = 0

    def display(self, message, level):
        if self.DISPLAY and (level <= self.DISPLAYLEVEL): print time.asctime() + " - " + message

    def assignPValRawOrg(self, lstMarker):
        calcRawPObj = calculationRawP.CalculationRawP(self.appConfig)
        loopcount    = 0
        amountMarker = len (lstMarker)
        for marker in lstMarker:
            loopcount += 1
            self.display ("Marker %i/%i" %(loopcount, amountMarker), 3)            
            self.display ("Calc. orig. raw p for marker line % i .." %marker['ln'],3)
            lstTvalPermut = self.inputDataFile.getLstPermutTvalues(marker['ln'], marker['id'])
            calcRawPObj.calcPValRawOrg(marker, lstTvalPermut)
            self.display ("Calc. orig. raw p for marker line % i finished." %marker['ln'],3)

    def sortMarkerByRawP(self, mrkListHndl):

        logFile_Sorting = 0
        if self.appConfig.logSorting == 1:
            logFile_Sorting = logfile.Logfile(self.appConfig.logFileSorting_Path, 1)
        
        sortedMarkerList = mrkListHndl.sortMarkerList(logFile_Sorting)                  
        return sortedMarkerList
            
        
    def calcRawPpermutations(self, marker):
             
        calcRawPObj = calculationRawP.CalculationRawP(self.appConfig)

        logFile_CalcRaw = 0
        if self.appConfig.logRawPCalc == 1:
            logFile_CalcRaw = logfile.Logfile(self.appConfig.logFileCalcRawP_Path, 1)
            logFile_CalcRaw.msg("Used Formula    : ")
            logFile_CalcRaw.msgEoln(self.appConfig.calcRawPFormula)
            logFile_CalcRaw.msg("Hyposthesis Test: ")
            logFile_CalcRaw.msgEoln(self.appConfig.testCharacter)                        
                   
        calcRawPObj.assignRawPvaluesPermutations(marker)
            
        if self.appConfig.logRawPCalc == 1:
            hlpFormat = formatHelper.FormatHelper()
            logFile_CalcRaw.msgEoln("==============================================================================")
            logFile_CalcRaw.msg("TestID:\t")
            logFile_CalcRaw.msgEoln(marker['id'])
            logFile_CalcRaw.msgEoln("")
            logFile_CalcRaw.msgEoln("\tOriginal\tPermut")
            logFile_CalcRaw.msgEoln("------------------------------------------------------------------------------")
            logFile_CalcRaw.msg("test stat.\t")
            logFile_CalcRaw.msg(hlpFormat.formatFloat(marker['tValOrg']))
            logFile_CalcRaw.msg("\t")
            logFile_CalcRaw.msgEoln(hlpFormat.formatFloatList(marker['tVectorPrm']))
            logFile_CalcRaw.msg("raw p     \t")
            logFile_CalcRaw.msg(hlpFormat.formatFloat(marker['pValRawOrg']))
            logFile_CalcRaw.msg("\t")
            logFile_CalcRaw.msgEoln(hlpFormat.formatFloatList(marker['lstPvalRawPrm']))
            logFile_CalcRaw.msgEoln("==============================================================================")
            logFile_CalcRaw.msgEoln("")
            logFile_CalcRaw.msgEoln("")

        # delete permutation t vector, to save memory
        marker['tVectorPrm'] = []


    def stepDownMinP(self, sortedMarkerList):
        
        logFile_StepDown = 0
        if self.appConfig.logStepDown == 1:
            logFile_StepDown = logfile.Logfile(self.appConfig.logFileStepDown_Path, 1)

        objCalcStepDownMinP = calculationStepDownMinP.CalculationStepDownMinP(self.appConfig)

        qVectorInitialize = 0
        qvectorMin = []
        qvectorSec = []        

        pvalueRawOld            = 10
        amount_permutations_old = -1
        loopcount = 0
        amountMarker = len(sortedMarkerList)
        for marker in sortedMarkerList:
            loopcount += 1
            self.display("Marker %i/%i" %(loopcount, amountMarker ),3)

            # ---------------------------------------------------------------------------------------------------
            # read marker data from file
            marker['tVectorPrm'] = self.inputDataFile.getLstPermutTvalues(marker['ln'], marker['id'])

            # ---------------------------------------------------------------------------------------------------
            # calculate raw p values for permutations            
            self.display("Calculating permutation raw p values marker %i (pval= %s) .." %(marker['ln'], marker['pValRawOrg']),3)
            lstPvalPermutations = self.calcRawPpermutations(marker)
            self.display("Calculating permutation raw p values marker %i finished." %marker['ln'],3)
            amount_permutations = len(marker['lstPvalRawPrm'])

            # ---------------------------------------------------------------------------------------------------
            # security checks
            # 1) order of original raw p values
            if marker['pValRawOrg'] > pvalueRawOld: sys.exit("order of raw pvalues wrong in step down procedure")
            pvalueRawOld = marker['pValRawOrg']
            # 2) check for same amount of permutations for every marker
            if (amount_permutations_old <> -1) and (amount_permutations <> amount_permutations_old):
                sys.exit("Irregular amount of test statistics for marker in line " + marker['ln'])
            amount_permutations_old = amount_permutations

            # ---------------------------------------------------------------------------------------------------
            # initialize the q vector, and q vector for second smallest raw p values            
            if qVectorInitialize == 0:
                qVectorInitialize = 1
                self.display("Initializing q vector ..", 3)
                for i in range(amount_permutations):
                    qvectorMin.append(1)
                if self.appConfig.calcGlobalPvalOption == 'EXTENDED':                    
                    qvectorSec = qvectorMin[:]
                self.display("Initializing q vector finished.",3)

            # ---------------------------------------------------------------------------------------------------
            # assign adjusted orig. p-value, calculate q vector of smallest and second smallest permut. p-values              
            self.display("Step Down Calculation marker %i .." %marker['ln'],3)
            qvector = objCalcStepDownMinP.performStepDownProcedure(marker, qvectorMin, qvectorSec, amount_permutations, logFile_StepDown)
            self.display("Step Down Calculation marker %i finished." %marker['ln'],3)
            

            # ---------------------------------------------------------------------------------------------------
            # save memory              
            marker['lstPvalRawPrm'] = []
                        
        return qvectorMin, qvectorSec, amount_permutations

    def enforceMonotonicity (self, markerListAdjSortedMaxToMinRawPval):

        objCalcStepDownMinP = calculationStepDownMinP.CalculationStepDownMinP(self.appConfig)

        logFile_StepDown = 0
        if self.appConfig.logStepDown == 1:
            logFile_StepDown = logfile.Logfile(self.appConfig.logFileStepDown_Path, 1)

        #--> consider that it must be already sorted from max to min (so it comes from step down proc)!!!            
        markerListAdj = objCalcStepDownMinP.enforceMonotonicity(markerListAdjSortedMaxToMinRawPval, logFile_StepDown)
         
        return markerListAdj

    def calculateGlobalPvalue (self, minQvector, secMinQVector, markerList):

        objCalcGlobal = calculationGlobalPvalue.CalculationGlobalPvalue(self.appConfig)
        logFile_Global = 0
        globalPvalMessage = ''
        if self.appConfig.logGlobalPvalCalc == 1:
            logFile_Global = logfile.Logfile(self.appConfig.logFileGlobal_Path, 1)

        if self.appConfig.calcGlobalPvalOption == 'SIMPLE':          
            globalPvalue = objCalcGlobal.calcGlobalP_simple (minQvector, markerList, logFile_Global)

        elif self.appConfig.calcGlobalPvalOption == 'EXTENDED':
            
            mrkListHnd = markerListHandler.MarkerListHandler(self.appConfig, markerList)            
            lstOrgRawPvalues = mrkListHnd.getLstRawPValOrg(logFile_Global)
            globalPvalue, globalPvalMessage = objCalcGlobal.calcGlobalP_extendSecondSmallest(lstOrgRawPvalues, minQvector, secMinQVector, logFile_Global)
            
        else:
            sys.exit("No valid option entered for calculation of global p value.")

        return globalPvalue, globalPvalMessage

    def writeResult (self, markerList, globalPvalue, globalPvalMessage, amountMarker, amount_permutations, startTimeStr, endTimeStr):

        resultFile = resultfile.Resultfile(self.appConfig)
        resultFile.writeFinalResult(markerList, globalPvalue, amountMarker, amount_permutations, startTimeStr, endTimeStr)
        
        return self.appConfig.resultFile_Path
        
    def run(self):        
        startTime = time.localtime()
        startTimeStr = "%i-%i-%i" %(startTime[3],startTime[4],startTime[5])
        
        if self.appConfig.useSplitFiles:
            self.display("Creating split files ..",0)
            totalLineCount = self.inputDataFile.createSplitFiles()
            self.display("Creating split files finished.",0)
            
        self.display("Reading Basis data ..",0)
        lstMarker, self.pValRawOrgIncluded = self.inputDataFile.getLstMarkerBasisInfo()
        self.display("Reading Basis data finished.",0)
        amountMarker = len (lstMarker)

        # id's unique?        
        lstTmpid = [""]
        for z in lstMarker:            
            if string.strip(z['id']) in lstTmpid:
                print 
                sys.exit ("--> Id '%s' not unique. See in line %i." %(z['id'],z['ln']))
            lstTmpid.append(string.strip(z['id']))
                

        # wrong number of markers given in the input file
        if amountMarker <= 1:
            print
            print "--> invalid number of provided tests."
            print "Please provide at least two tests."
            sys.exit("Program terminated")

        # choose calculation approach, either Ge et al. or Becker and Knapp
        if self.pValRawOrgIncluded == 0:
            self.appConfig.calcRawPFormula = "BECKER"
            self.display("Calculating original raw p values ..",0)
            self.assignPValRawOrg(lstMarker)
            self.display("Calculating original raw p values finished.",0)
        else:
            self.appConfig.calcRawPFormula = "GE_ET_AL"

        mrkListHndl = markerListHandler.MarkerListHandler(self.appConfig, lstMarker)
        self.display("Sorting by raw p value ..",0)
        sortedMarkerList = self.sortMarkerByRawP(mrkListHndl)
        mrkListHndl.lstMarkerList = sortedMarkerList
        self.display("Sorting by raw p value finished.",0)

        self.display("Step Down Calculation ..",0)
        minQVector, secMinQVector, amount_permutations = self.stepDownMinP(sortedMarkerList)
        self.display("Step Down Calculation finished.",0)
        
        #--> consider that markerListAdj is already sorted from max to min
        #   (so it comes from step down proc)
        #   this is the necessary sorting for the next step
        self.display("Enforcing Monotonicity ..",0)
        self.enforceMonotonicity(sortedMarkerList)
        self.display("Enforcing Monotonicity finished.",0)
        
        self.display("Calculating Global p value ..",0)
        globalPval, globalPvalMessage = self.calculateGlobalPvalue(minQVector, secMinQVector, sortedMarkerList)
        self.display("Calculating Global p value finished.",0)

        endTime = time.localtime()
        endTimeStr   = "%i-%i-%i" %(endTime[3],endTime[4],endTime[5])
        self.display("Writing Result ..",0)
        self.writeResult(sortedMarkerList, globalPval, globalPvalMessage, amountMarker, amount_permutations, startTimeStr, endTimeStr)        
        self.display("Writing Result finished.",0)
        
        if self.LOGTIME:
            endTime = time.localtime()            
            endTimeStr   = "%i-%i-%i" %(endTime[3],endTime[4],endTime[5])
            s = ''
            timeLog = open(self.logtimePath , 'a')
            
            #timeLog.write('---------------------------\n')
            msg     = '%s\t%i\t%i\t%s\t%s\t%s\n' %(os.path.basename(self.appConfig.dataInputFile_Path), amountMarker, amount_permutations, startTimeStr, endTimeStr, s)
            timeLog.write(msg)
            timeLog.close()

        if self.PROMPTRESULT:
            for m in sortedMarkerList:
                separator=  '------------------------------------------------------'
                markerID = 'Test-ID : ' +  str(m['id'])
                pvalOrg  = 'PvalOrg   : ' +  str(m['pValRawOrg'])
                pvalAdj  = 'PvalAdj   : ' +  str(m['adjPval'])
                pvalAdjMonotonicity = 'PvalAdjMon: ' +  str(m['adjPval_enforcedMonotonicity'])
        
                print("")
                print(separator)
                print(markerID)
                print(pvalOrg)
                print(pvalAdj)
                print(pvalAdjMonotonicity)

            print
            print 'Global Adjusted minP value: ' + str(globalPval)
            print
            if self.pValRawOrgIncluded:
                msg = "provided in data file"
            else:
                msg = "calculated"

            print ""            
            print "Raw p-values          :\t%s" %msg
            print "TestCharacter         :\t%s" %self.appConfig.testCharacter          
            print "RawP - Formula        :\t%s" %self.appConfig.calcRawPFormula
            print "Global pval - Formula :\t%s" %self.appConfig.calcGlobalPvalOption
            print 
            print "Number of marker               :\t%i" %amountMarker
            print "Number of perm. test statistics:\t%i" %amount_permutations
            print            
            
            print"InputFile:\t%s" %self.appConfig.dataInputFile_Path

            print "Start-Time:\t%s" %startTimeStr
            print "End  -Time:\t%s" %endTimeStr
            print
        
            print "Result  results stored in "
            print self.appConfig.resultFile_Path

       

            
            
            
